local Addon, private = ...

-- Builtins
local _G = _G
local math = math
local pairs = pairs
local pcall = pcall
local print = print
local string = string
local table = table
local tonumber = tonumber
local tostring = tostring

-- Globals
local Command = Command
local Event = Event
local Inspect = Inspect
local UI = UI
local UIParent = UIParent
local Utility = Utility
local zlib = zlib

-- Locals
local padding = 5
local labelFormats = {
	player = "Player Storage:\n(%i/%i used)",
	guild = "Guild Storage:\n(%i/%i used)",
	size = "Size: %i",
	checksum = "Checksum: %s",
	uncompressed = "Uncompressed: %i",
	compressed = "Compressed: %i",
	ratio = "Ratio: %i%%",
}

-- Prototypes
local identifierSelected
local dataOrIdentifierChanged
local segmentChanged
local getList, getListCallback, listReceived
local clearData, clearDataCallback
local getData, getDataCallback, dataReceived
local sendData, sendDataCallback

setfenv(1, private)

if(Addon.toc.debug) then
	log = print
else
	log = function() end
end

-- UI creation methods
-- ============================================================================

local function createTextField(parent, focusHooks)
	local background = UI.CreateFrame("Frame", "", parent)
	local self = UI.CreateFrame("RiftTextfield", "", background)
	self:SetBackgroundColor(0, 0, 0, 0.7)
	
	background:SetBackgroundColor(1, 1, 1, 0.8)
	background:SetPoint("TOPLEFT", self, "TOPLEFT", -1, -1)
	background:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", 1, 1)
	
	if(focusHooks) then
		function self.Event:KeyFocusGain()
			self:SetSelection(0, string.len(self:GetText()))
		end

		function self.Event:KeyFocusLoss()
			if(self:GetText() == "") then
				self:SetText(Inspect.Unit.Detail("player").name)
			end
		end
	end

	self:SetText("")
	return self
end

local function createDataTextField(parent)
	local self = createTextField(parent)

	function self.Event:KeyDown(button)
		local code = string.byte(button)
		local txt = self:GetText()
		local pos = self:GetCursor()
		self:SetKeyFocus(true) 
		if tonumber(code) == 13 then
			self:SetText(string.insert(txt, pos + 1, "\n"))
			self:SetCursor(pos + 1)
		elseif tonumber(code) == 9 then
			self:SetText(string.insert(txt, pos + 1, "\t"))
			self:SetCursor(pos + 1)
		end
	end

	function self.Event:TextfieldChange()
		parent.realText = self:GetText()
		dataOrIdentifierChanged(parent, self)
	end

	return self
end

local function createCheckbox(parent, text, checked)
	local self = UI.CreateFrame("RiftCheckbox", "", parent)
	self.label = UI.CreateFrame("Text", "", self)
	self.label:SetPoint("LEFTCENTER", self, "RIGHTCENTER")
	self.label:SetText(text)
	self:SetChecked(checked)
	return self
end

local function createIdentifierList(parent, width, entries)
	local self = UI.CreateFrame("Frame", "", parent)
	self:SetBackgroundColor(0, 0, 0)
	self.selected = ""
	
	local scrollbar = UI.CreateFrame("RiftScrollbar", "", self)
	scrollbar:SetPoint("TOPLEFT", self, "TOPRIGHT")
	scrollbar:SetPoint("BOTTOMLEFT", self, "BOTTOMRIGHT")
	scrollbar.Event.ScrollbarChange = function() self:Update() end
	
	local label = UI.CreateFrame("Text", "", self)
	label:SetText("X")
	label:SetPoint("TOPLEFT", self, "TOPLEFT")
	label:SetPoint("TOPRIGHT", scrollbar, "TOPLEFT")
	label.Event.MouseIn = function(label) label:SetBackgroundColor(0.3, 0.3, 0.3) end
	label.Event.MouseOut = function(label) if(label:GetText() ~= self.selected) then label:SetBackgroundColor(0, 0, 0) end end
	label.Event.LeftDown = function(label) self.selected = label:GetText() identifierSelected(parent, self, label:GetText()) self:Update() end
	self.labels = { label }
	
	for i = 2, entries do
		local label = UI.CreateFrame("Text", "", self)
		label:SetText("X")
		label:SetPoint("TOPLEFT", self.labels[i - 1], "BOTTOMLEFT")
		label:SetPoint("TOPRIGHT", self.labels[i - 1], "BOTTOMRIGHT")
		label.Event.MouseIn = function(label) label:SetBackgroundColor(0.3, 0.3, 0.3) end
		label.Event.MouseOut = function(label) if(label:GetText() ~= self.selected) then label:SetBackgroundColor(0, 0, 0) end end
		label.Event.LeftDown = function(label) self.selected = label:GetText() identifierSelected(parent, self, label:GetText()) self:Update() end
		self.labels[i] = label
	end
	
	self:SetHeight(self.labels[1]:GetHeight() * entries)
	self:SetWidth(width - scrollbar:GetWidth())
	
	function self:SetIdentifiers(identifiers)
		self.identifiers = identifiers
		scrollbar:SetRange(0, math.max(1, #identifiers - entries))
		scrollbar:SetEnabled(#identifiers > entries)
		scrollbar:SetPosition(0)
		self.selected = ""
		self:Update()
	end
	
	function self:Update()
		local offset = math.floor(scrollbar:GetPosition())
		for i = 1, entries do
			local index = i + offset
			if(index > #self.identifiers) then
				self.labels[i]:SetVisible(false)
			else
				self.labels[i]:SetVisible(true)
				self.labels[i]:SetText(self.identifiers[index])
				if(self.identifiers[index] == self.selected) then
					self.labels[i]:SetBackgroundColor(0.3, 0.3, 0.3)
				else
					self.labels[i]:SetBackgroundColor(0, 0, 0)
				end
			end
		end
	end
	
	return self
end

function identifierSelected(self, list, identifier)
	if(list == self.playerList) then
		self.checksum:SetText(string.format(labelFormats.checksum, tostring(self.player[identifier])))
		self.guildList.selected = ""
		self.guildList:Update()
		self.segment = "player"
		self.identifier = identifier
	elseif(list == self.guildList) then
		self.checksum:SetText(string.format(labelFormats.checksum, tostring(self.guild[identifier])))
		self.playerList.selected = ""
		self.playerList:Update()
		self.segment = "guild"
		self.identifier = identifier
	end
	self.loadBtn:SetEnabled(identifier ~= "")
	self.deleteBtn:SetEnabled(identifier ~= "")
end

function dataOrIdentifierChanged(self)
	local size = Utility.Storage.Size(self.loadedIdentifier:GetText(), self.realText)
	self.loadedSize:SetText(string.format(labelFormats.size, size))
	self.loadedChecksum:SetText(string.format(labelFormats.checksum, Utility.Storage.Checksum(self.realText)))
	local used, max = Inspect.Storage.Used(self.checkPlayer:GetChecked() and "player" or "guild")
	local free = max - used
	
	if(size <= free) then
		self.loadedSize:SetFontColor(0, 1, 0) -- Green: fits in unused space
	elseif(size > max) then
		self.loadedSize:SetFontColor(1, 0.3, 0) -- Red: won't ever fit
	else
		self.loadedSize:SetFontColor(1, 0.8, 0) -- Yellow: does not fit in unused space but might fit if data is replaced
	end
end

function segmentChanged(self)
	dataOrIdentifierChanged(self) -- Update size color as it is affected by segment size
	self.playerPermissions:SetVisible(self.checkPlayer:GetChecked())
	self.guildPermissions:SetVisible(self.checkGuild:GetChecked())
end

local function createWindow()
	if(not Inspect.System.Secure()) then
		Command.System.Watchdog.Quiet()
	end
	
	local context = UI.CreateContext("StorageViewer")
	local self = UI.CreateFrame("RiftWindow", "", context)
	self:SetTitle(Addon.toc.Name)
	
	local label
	
	self:SetController("content")
	self:SetWidth(900)
	self:SetHeight(500)
	self:SetPoint("TOPLEFT", UIParent, "TOPLEFT", (UIParent:GetWidth() - self:GetWidth()) / 2, (UIParent:GetHeight() - self:GetHeight()) / 2)
	
	local btn = UI.CreateFrame("RiftButton", "", self)
	btn:SetSkin("close")
	btn:SetPoint("TOPRIGHT", self:GetBorder(), "TOPRIGHT", -8, 16)
	btn.Event.LeftPress = function() self:SetVisible(false) end

	-- Target textfield and list button
	self.target = createTextField(self, true)
	self.target:SetPoint("TOPLEFT", self, "TOPLEFT", padding, padding)
	self.target:SetText(Inspect.Unit.Detail("player").name)
	self.target:SetWidth(135)
	
	local listBtn = UI.CreateFrame("RiftButton", "", self)
	listBtn:SetPoint("BOTTOMLEFT", self.target, "BOTTOMRIGHT", 0, 6)
	listBtn:SetText("List")
	listBtn.Event.LeftPress = getList
	
	self.playerLabel = UI.CreateFrame("Text", "", self)
	self.playerLabel:SetText(string.format(labelFormats.player, Inspect.Storage.Used("player")))
	self.playerLabel:SetPoint("TOPLEFT", self.target, "BOTTOMLEFT", 0, padding)
	self.playerLabel:SetPoint("TOPRIGHT", self.target, "BOTTOMRIGHT", 0, padding)
	
	-- Player storage list
	self.playerList = createIdentifierList(self, 135, 20)
	self.playerList:SetPoint("TOPLEFT", self.playerLabel, "BOTTOMLEFT")
	self.playerList:SetIdentifiers({ })
	
	self.guildLabel = UI.CreateFrame("Text", "", self)
	self.guildLabel:SetText(string.format(labelFormats.guild, Inspect.Storage.Used("guild")))
	self.guildLabel:SetPoint("TOPLEFT", listBtn, "BOTTOMLEFT", 0, padding)
	self.guildLabel:SetPoint("TOPRIGHT", listBtn, "BOTTOMRIGHT", 0, padding - 6)
	
	-- Guild storage list
	self.guildList = createIdentifierList(self, 135, 20)
	self.guildList:SetPoint("TOPLEFT", self.guildLabel, "BOTTOMLEFT")
	self.guildList:SetIdentifiers({ })
	
	-- Load and Delete buttons
	self.loadBtn = UI.CreateFrame("RiftButton", "", self)
	self.loadBtn:SetPoint("TOPLEFT", self.playerList, "BOTTOMLEFT")
	self.loadBtn:SetText("Load")
	self.loadBtn:SetEnabled(false)
	self.loadBtn.Event.LeftPress = getData
	
	self.deleteBtn = UI.CreateFrame("RiftButton", "", self)
	self.deleteBtn:SetPoint("TOPLEFT", self.loadBtn, "TOPRIGHT")
	self.deleteBtn:SetText("Delete")
	self.deleteBtn:SetEnabled(false)
	self.deleteBtn.Event.LeftPress = clearData
	
	-- Checksum label
	self.checksum = UI.CreateFrame("Text", "", self)
	self.checksum:SetPoint("TOPLEFT", self.loadBtn, "BOTTOMLEFT")
	self.checksum:SetText(" ")
	
	-- Loaded datum details
	self.loadedTarget = UI.CreateFrame("Text", "", self)
	self.loadedTarget:SetPoint("TOPLEFT", self, "TOPRIGHT", -padding - 135, padding)
	self.loadedTarget:SetFontSize(16)
	self.loadedTarget:SetText(" ")
	self.loadedTarget:SetFontColor(1, 1, 0)
	
	self.loadedIdentifier = createTextField(self)
	self.loadedIdentifier:SetPoint("TOPLEFT", self.loadedTarget, "BOTTOMLEFT", 0, padding)
	self.loadedIdentifier:SetWidth(135)
	self.loadedIdentifier.Event.TextfieldChange = function() dataOrIdentifierChanged(self) end
	
	self.checkPlayer = createCheckbox(self, "Player", true)
	self.checkPlayer:SetPoint("TOPLEFT", self.loadedIdentifier, "BOTTOMLEFT", 0, padding)
	
	self.checkGuild = createCheckbox(self, "Guild", false)
	self.checkGuild:SetPoint("LEFTCENTER", self.checkPlayer.label, "RIGHTCENTER", 10, 0)
	
	self.checkPlayer.Event.CheckboxChange = function() self.checkGuild:SetChecked(not self.checkPlayer:GetChecked()) segmentChanged(self) end
	self.checkGuild.Event.CheckboxChange = function() self.checkPlayer:SetChecked(not self.checkGuild:GetChecked()) segmentChanged(self) end
	
	self.playerPermissions = UI.CreateFrame("Frame", "", self)
	self.playerPermissions:SetBackgroundColor(0, 0, 0, 0.5)
	self.playerPermissions:SetPoint("TOPLEFT", self.checkPlayer, "BOTTOMLEFT", 0, padding)
	self.playerPermissions:SetWidth(135)
	label = UI.CreateFrame("Text", "", self.playerPermissions)
	label:SetPoint("TOPLEFT", self.playerPermissions, "TOPLEFT", padding, 0)
	label:SetText("Read Permissions")
	label:SetFontSize(14)
	label:SetFontColor(1, 1, 0)
	self.playerReadPublic = createCheckbox(self.playerPermissions, "Public", true)
	self.playerReadPublic:SetPoint("TOPLEFT", label, "BOTTOMLEFT", padding, padding)
	self.playerReadGuild = createCheckbox(self.playerPermissions, "Guild", false)
	self.playerReadGuild:SetPoint("TOPLEFT", self.playerReadPublic, "BOTTOMLEFT")
	self.playerReadPrivate = createCheckbox(self.playerPermissions, "Private", false)
	self.playerReadPrivate:SetPoint("TOPLEFT", self.playerReadGuild, "BOTTOMLEFT")
	local playerReadChanging = false
	self.playerReadPublic.Event.CheckboxChange = function(cb)
		if(not playerReadChanging) then
			playerReadChanging = true
			if(not cb:GetChecked()) then cb:SetChecked(true) else 
				self.playerReadPublic:SetChecked(false)
				self.playerReadGuild:SetChecked(false)
				self.playerReadPrivate:SetChecked(false)
				cb:SetChecked(true)
			end
			playerReadChanging = false
		end
	end
	self.playerReadGuild.Event.CheckboxChange = self.playerReadPublic.Event.CheckboxChange
	self.playerReadPrivate.Event.CheckboxChange = self.playerReadPublic.Event.CheckboxChange
	self.playerPermissions:SetHeight(self.playerReadPrivate:GetBottom() - self.playerPermissions:GetTop() + padding)
	
	self.guildPermissions = UI.CreateFrame("Frame", "", self)
	self.guildPermissions:SetBackgroundColor(0, 0, 0, 0.5)
	self.guildPermissions:SetPoint("TOPLEFT", self.checkPlayer, "BOTTOMLEFT", 0, padding)
	self.guildPermissions:SetWidth(135)
	label = UI.CreateFrame("Text", "", self.guildPermissions)
	label:SetPoint("TOPLEFT", self.guildPermissions, "TOPLEFT", padding, 0)
	label:SetText("Read Permission")
	label:SetFontSize(14)
	label:SetFontColor(1, 1, 0)
	self.guildReadPublic = createCheckbox(self.guildPermissions, "Public", true)
	self.guildReadPublic:SetPoint("TOPLEFT", label, "BOTTOMLEFT", padding, padding)
	self.guildReadGuild = createCheckbox(self.guildPermissions, "Guild", false)
	self.guildReadGuild:SetPoint("TOPLEFT", self.guildReadPublic, "BOTTOMLEFT")
	self.guildReadOfficer = createCheckbox(self.guildPermissions, "Officer", false)
	self.guildReadOfficer:SetPoint("TOPLEFT", self.guildReadGuild, "BOTTOMLEFT")
	local guildReadChanging = false
	self.guildReadPublic.Event.CheckboxChange = function(cb)
		if(not guildReadChanging) then
			guildReadChanging = true
			if(not cb:GetChecked()) then cb:SetChecked(true) else 
				self.guildReadPublic:SetChecked(false)
				self.guildReadGuild:SetChecked(false)
				self.guildReadOfficer:SetChecked(false)
				cb:SetChecked(true)
				if(self.guildReadOfficer:GetChecked()) then
					self.guildWriteGuild:SetChecked(false)
					self.guildWriteOfficer:SetChecked(true)
				end
			end
			guildReadChanging = false
		end
	end
	self.guildReadGuild.Event.CheckboxChange = self.guildReadPublic.Event.CheckboxChange
	self.guildReadOfficer.Event.CheckboxChange = self.guildReadPublic.Event.CheckboxChange
	label = UI.CreateFrame("Text", "", self.guildPermissions)
	label:SetPoint("TOPLEFT", self.guildReadOfficer, "BOTTOMLEFT", -padding, padding)
	label:SetText("Write Permission")
	label:SetFontSize(14)
	label:SetFontColor(1, 1, 0)
	self.guildWriteGuild = createCheckbox(self.guildPermissions, "Guild", true)
	self.guildWriteGuild:SetPoint("TOPLEFT", label, "BOTTOMLEFT", padding, padding)
	self.guildWriteOfficer = createCheckbox(self.guildPermissions, "Officer", false)
	self.guildWriteOfficer:SetPoint("TOPLEFT", self.guildWriteGuild, "BOTTOMLEFT")
	self.guildWriteGuild.Event.CheckboxChange = function() self.guildWriteOfficer:SetChecked(not self.guildWriteGuild:GetChecked()) end
	self.guildWriteOfficer.Event.CheckboxChange = function() self.guildWriteGuild:SetChecked(not self.guildWriteOfficer:GetChecked()) end
	self.guildPermissions:SetHeight(self.guildWriteOfficer:GetBottom() - self.guildPermissions:GetTop() + padding)
	self.guildPermissions:SetVisible(false)
	
	-- Compression
	local compression = UI.CreateFrame("Frame", "", self)
	compression:SetBackgroundColor(0, 0, 0, 0.5)
	compression:SetPoint("TOPLEFT", self.guildPermissions, "BOTTOMLEFT", 0, padding)
	compression:SetPoint("TOPRIGHT", self.guildPermissions, "BOTTOMRIGHT", 0, padding)
	
	label = UI.CreateFrame("Text", "", compression)
	label:SetPoint("TOPLEFT", compression, "TOPLEFT")
	label:SetFontSize(14)
	label:SetFontColor(1, 1, 0)
	label:SetText("Compression")
	
	self.uncompressed = UI.CreateFrame("Text", "", compression)
	self.uncompressed:SetText(string.format(labelFormats.uncompressed, 0))
	self.uncompressed:SetPoint("TOPLEFT", label, "BOTTOMLEFT", padding, 0)
	
	self.compressed = UI.CreateFrame("Text", "", compression)
	self.compressed:SetText(string.format(labelFormats.compressed, 0))
	self.compressed:SetPoint("TOPLEFT", self.uncompressed, "BOTTOMLEFT")
	
	self.ratio = UI.CreateFrame("Text", "", compression)
	self.ratio:SetText(string.format(labelFormats.ratio, 0))
	self.ratio:SetPoint("TOPLEFT", self.compressed, "BOTTOMLEFT")
	
	label = UI.CreateFrame("Text", "", compression)
	label:SetPoint("TOPLEFT", self.ratio, "BOTTOMLEFT")
	label:SetText(string.format("Compr. Level (%i-%i)", zlib.BEST_SPEED, zlib.BEST_COMPRESSION))
	local level = createTextField(compression, false)
	level:SetPoint("TOPRIGHT", self.ratio, "BOTTOMLEFT", 135 - padding - 1, 0)
	level:SetWidth(20)
	level:SetText("6")
	level.Event.KeyFocusLoss = function()
		local value = tonumber(level:GetText())
		if(value) then
			value = math.min(zlib.BEST_COMPRESSION, math.max(zlib.BEST_SPEED, value))
		else
			value = 6
		end
		level:SetText(tostring(value))
	end
	
	local deflateBtn = UI.CreateFrame("RiftButton", "", compression)
	deflateBtn:SetText("Compress")
	deflateBtn:SetPoint("TOPLEFT", label, "BOTTOMLEFT", -padding, 0)
	deflateBtn.Event.LeftPress = function()
		local identifier = self.loadedIdentifier:GetText()
		local size1 = Utility.Storage.Size(identifier, self.realText)
		self.realText = zlib.deflate()(self.realText, "finish")
		self.text:SetText(self.realText)
		local size2 = Utility.Storage.Size(identifier, self.realText)
		self.uncompressed:SetText(string.format(labelFormats.uncompressed, size1))
		self.compressed:SetText(string.format(labelFormats.compressed, size2))
		self.ratio:SetText(string.format(labelFormats.ratio, math.floor(size2 / size1 * 100)))
		dataOrIdentifierChanged(self)
	end
	
	local inflateBtn = UI.CreateFrame("RiftButton", "", compression)
	inflateBtn:SetText("Uncompress")
	inflateBtn:SetPoint("TOPLEFT", deflateBtn, "BOTTOMLEFT")
	inflateBtn.Event.LeftPress = function()
		local identifier = self.loadedIdentifier:GetText()
		local size1 = Utility.Storage.Size(identifier, self.realText)
		self.realText = zlib.inflate()(self.realText, "finish")
		self.text:SetText(self.realText)
		local size2 = Utility.Storage.Size(identifier, self.realText)
		self.compressed:SetText(string.format(labelFormats.compressed, size1))
		self.uncompressed:SetText(string.format(labelFormats.uncompressed, size2))
		self.ratio:SetText(string.format(labelFormats.ratio, math.floor(size1 / size2 * 100)))
		dataOrIdentifierChanged(self)
	end
	
	compression:SetHeight(inflateBtn:GetBottom() - compression:GetTop() + padding)
	
	-- Text edit field
	self.text = createDataTextField(self)
	self.realText = ""
	self.text:SetText(self.realText)
	self.text:SetPoint("TOPLEFT", listBtn, "TOPRIGHT", padding, 0)
	self.text:SetPoint("TOPRIGHT", self.loadedTarget, "TOPLEFT", -padding, 0)
	self.text:SetPoint("BOTTOMLEFT", self.deleteBtn, "BOTTOMRIGHT", padding, 0)
	
	self.loadedSize = UI.CreateFrame("Text", "", self)
	self.loadedSize:SetText(string.format(labelFormats.size, 0))
	self.loadedSize:SetPoint("TOPLEFT", self.text, "BOTTOMLEFT")
	
	self.loadedChecksum = UI.CreateFrame("Text", "", self)
	self.loadedChecksum:SetText(string.format(labelFormats.checksum, ""))
	self.loadedChecksum:SetPoint("LEFTCENTER", self.loadedSize, "LEFTCENTER", 70, 0)
--[[	
	-- Serialization
	local serialization = UI.CreateFrame("Frame", "", self)
	serialization:SetBackgroundColor(0, 0, 0, 0.5)
	serialization:SetPoint("TOPLEFT", compression, "BOTTOMLEFT", 0, padding)
	serialization:SetPoint("TOPRIGHT", compression, "BOTTOMRIGHT", 0, padding)
	
	label = UI.CreateFrame("Text", "", serialization)
	label:SetPoint("TOPLEFT", serialization, "TOPLEFT")
	label:SetFontSize(14)
	label:SetFontColor(1, 1, 0)
	label:SetText("De/Serialize Global")
	
	local serializeText = createTextField(serialization)
	serializeText:SetPoint("TOPLEFT", label, "BOTTOMLEFT", padding, 0)
	serializeText:SetWidth(135 - 2 * padding)
	serializeText:SetText("<global symbol>")
	serializeText.Event.KeyFocusGain = function(self) if(self:GetText() == "<global symbol>") then self:SetText("") end end
	serializeText.Event.KeyFocusLoss = function(self) if(self:GetText() == "") then self:SetText("<global symbol>") end end
	
	local serializeBtn = UI.CreateFrame("RiftButton", "", serialization)
	serializeBtn:SetText("Serialize")
	serializeBtn:SetPoint("TOPLEFT", serializeText, "BOTTOMLEFT", -padding, 0)
	serializeBtn.Event.LeftPress = function()
		local serialized = Utility.Serialize.Inline(_G[serializeText:GetText()])
		serializeText:SetText("{2,3,a=1,b=5}")
		self.text:SetText("{{2,3,a=1,b=5}")
	end
	
	local deserializeBtn = UI.CreateFrame("RiftButton", "", serialization)
	deserializeBtn:SetText("Deserialize")
	deserializeBtn:SetPoint("TOPLEFT", serializeBtn, "BOTTOMLEFT")
	
	serialization:SetHeight(deserializeBtn:GetBottom() - serialization:GetTop() + padding)
]]	
	-- THE BUTTON
	local sendBtn = UI.CreateFrame("RiftButton", "", self)
	sendBtn:SetText("Upload")
	sendBtn:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", 0, 0)
	sendBtn.Event.LeftPress = sendData
	
	table.insert(Event.Storage.List, { listReceived, Addon.identifier, "listReceived" })
	table.insert(Event.Storage.Get, { dataReceived, Addon.identifier, "dataReceived" })
	
	-- Make the window movable
	self:GetContent().Event.LeftDown = function()
		local mouse = Inspect.Mouse()
		self.mouseOffsetX, self.mouseOffsetY = self:GetLeft() - mouse.x, self:GetTop() - mouse.y
	end
	self:GetContent().Event.LeftUp = function()
		self.mouseOffsetX, self.mouseOffsetY = nil, nil
	end
	self:GetContent().Event.MouseMove = function()
		if(self.mouseOffsetX) then
			local mouse = Inspect.Mouse()
			local x = mouse.x + self.mouseOffsetX
			local y = mouse.y + self.mouseOffsetY
			self:SetPoint("TOPLEFT", UIParent, "TOPLEFT", x, y)
		end
	end
	self:GetBorder().Event.LeftDown = self:GetContent().Event.LeftDown
	self:GetBorder().Event.LeftUp = self:GetContent().Event.LeftUp
	self:GetBorder().Event.MouseMove = self:GetContent().Event.MouseMove
	
	self.waitingForTarget = ""
	self.waitingForSegment = ""
	self.waitingForIdentifier = ""
	self.waitingForList = 0
	MainWindow = self
end

-- Storage processing methods
-- ============================================================================

function getList()
	MainWindow.waitingForList = 2
	local target = MainWindow.target:GetText()
	local success, result = pcall(Inspect.Unit.Detail, target)
	if(success) then
		target = result.name
	end
	local success1 = pcall(Command.Storage.List, target, "player", function (...) getListCallback(target..":player", ...) end)
	local success2 = pcall(Command.Storage.List, target, "guild", function (...) getListCallback(target..":guild", ...) end)
	if(not (success1 and success2)) then
		print("Failed listing storage for " .. target)
		MainWindow.waitingForList = 0
	else
		MainWindow.target:SetText(target)
	end
end

function getListCallback(origin, failure, message)
	if(failure) then
		MainWindow.waitingForList = MainWindow.waitingForList - 1
		print("Failed listing storage for " .. origin .. " - ", message or "")
	end
end

function listReceived(target, segment, identifiers)
	if(MainWindow.waitingForList > 0) then
		MainWindow.waitingForList = MainWindow.waitingForList - 1
		MainWindow[segment] = identifiers
		local names = { }
		for k, v in pairs(identifiers) do
			names[#names + 1] = k
		end
		table.sort(names)
		MainWindow.checksum:SetText("")
		MainWindow.loadBtn:SetEnabled(false)
		MainWindow.deleteBtn:SetEnabled(false)
		MainWindow.identifier = ""
		MainWindow[segment .. "List"]:SetIdentifiers(names)
		MainWindow[segment .. "Label"]:SetText(string.format(labelFormats[segment], Inspect.Storage.Used(segment)))
	end
end

function clearData()
	if(MainWindow.segment and MainWindow.identifier ~= "") then
		local temp = MainWindow.segment .. ":" .. MainWindow.identifier
		if(not pcall(Command.Storage.Clear, MainWindow.segment, MainWindow.identifier, function(...) clearDataCallback(temp, ...) end)) then
			print("Failed clearing storage element" .. temp)
		end
	end
end

function clearDataCallback(origin, failure, message)
	if(failure) then
		print("Failed clearing storage element " .. origin .. " - ", message or "")
	end
	getList()
end

function getData()
	log(MainWindow.target:GetText(), MainWindow.segment, MainWindow.identifier)
	MainWindow.waitingForTarget = MainWindow.target:GetText()
	MainWindow.waitingForSegment = MainWindow.segment
	MainWindow.waitingForIdentifier = MainWindow.identifier
	local temp = MainWindow.target:GetText() .. ":" .. MainWindow.segment .. ":" .. MainWindow.identifier
	local success = pcall(Command.Storage.Get, MainWindow.target:GetText(), MainWindow.segment, MainWindow.identifier, function(...) getDataCallback(temp, ...) end)
	if(not success) then
		print("Failed getting storage element " .. temp)
		MainWindow.waitingForTarget = ""
		MainWindow.waitingForSegment = ""
		MainWindow.waitingForIdentifier = ""
	end
end

function getDataCallback(origin, failure, message)
	if(failure) then
		print("Failed getting storage element " .. origin .. " - ", message or "")
		MainWindow.waitingForSegment = ""
		MainWindow.waitingForIdentifier = ""
	end
end

function dataReceived(target, segment, identifier, read, write, data)
	local self = MainWindow
	log(target, segment, identifier, read, write, data)
	if(target == self.waitingForTarget and segment == self.waitingForSegment and identifier == self.waitingForIdentifier) then
		self.waitingForTarget = ""
		self.waitingForSegment = ""
		self.waitingForIdentifier = ""
		
		self.loadedTarget:SetText(target)
		self.loadedIdentifier:SetText(identifier)
		self.realText = data
		self.text:SetText(data)
		self.checkPlayer:SetChecked(segment == "player")
		self.checkGuild:SetChecked(segment == "guild")
		
		if(segment == "player") then
			if(read == "public") then
				self.playerReadPublic:SetChecked(true)
			elseif(read == "guild") then
				self.playerReadGuild:SetChecked(true)
			elseif(read == "private") then
				self.playerReadPrivate:SetChecked(true)
			end
		else
			if(read == "public") then
				self.guildReadPublic:SetChecked(true)
			elseif(read == "guild") then
				self.guildReadGuild:SetChecked(true)
			elseif(read == "officer") then
				self.guildReadOfficer:SetChecked(true)
			end
			self.guildWriteGuild:SetChecked(write == "guild")
			self.guildWriteOfficer:SetChecked(write == "officer")
		end
		
		self.uncompressed:SetText(string.format(labelFormats.uncompressed, 0))
		self.compressed:SetText(string.format(labelFormats.compressed, 0))
		self.ratio:SetText(string.format(labelFormats.ratio, 0))
		
		dataOrIdentifierChanged(self)
	end
end

function sendData()
	local self = MainWindow
	local segment = self.checkPlayer:GetChecked() and "player" or "guild"
	local identifier = self.loadedIdentifier:GetText()
	local read, write
	if(segment == "player") then
		read = (self.playerReadPublic:GetChecked() and "public") or
			(self.playerReadGuild:GetChecked() and "guild") or
			(self.playerReadPrivate:GetChecked() and "private")
		write = "private"
	else
		read = (self.guildReadPublic:GetChecked() and "public") or
			(self.guildReadGuild:GetChecked() and "guild") or
			(self.guildReadOfficer:GetChecked() and "officer")
		write = self.guildWriteGuild:GetChecked() and "guild" or "officer"
	end
	log(segment, identifier, read, write, self.realText)
	local temp = segment .. ":" .. identifier
	local success = pcall(Command.Storage.Set, segment, identifier, read, write, self.realText, function(...) sendDataCallback(temp, ...) end)
	if(not success) then
		print("Error setting storage element " .. temp)
	end
end

function sendDataCallback(origin, failure, message)
	if(failure) then
		print("Failed setting storage element " .. origin, message or "")
	end
	MainWindow.loadedTarget:SetText(Inspect.Unit.Detail("player").name)
	getList()
end

local function toggleWindow()
	if(not MainWindow) then
		createWindow()
	else
		MainWindow:SetVisible(not MainWindow:GetVisible())
	end
end

table.insert(Command.Slash.Register(Addon.identifier), { toggleWindow, Addon.identifier, "toggleWindow" })
table.insert(Command.Slash.Register("sv"), { toggleWindow, Addon.identifier, "toggleWindow" })
